@@ -52,15 +52,12 @@ static const struct
 gcry_mpi_t
 pk_ecdh_default_params (unsigned int qbits)
 {
-  byte *kek_params;
+  byte kek_params[4] = {
+    3, /* Number of bytes to follow. */
+    1  /* Version for KDF+AESWRAP.   */
+  };
   int i;
 
-  kek_params = xtrymalloc (4);
-  if (!kek_params)
-    return NULL;
-  kek_params[0] = 3; /* Number of bytes to follow. */
-  kek_params[1] = 1; /* Version for KDF+AESWRAP.   */
-
   /* Search for matching KEK parameter.  Defaults to the strongest
      possible choices.  Performance is not an issue here, only
      interoperability.  */
@@ -78,7 +75,7 @@ pk_ecdh_default_params (unsigned int qbits)
   if (DBG_CRYPTO)
     log_printhex (kek_params, sizeof(kek_params), "ECDH KEK params are");
 
-  return gcry_mpi_set_opaque (NULL, kek_params, 4 * 8);
+  return gcry_mpi_set_opaque_copy (NULL, kek_params, 4 * 8);
 }
 
 
@@ -159,11 +156,11 @@ build_kdf_params (unsigned char kdf_params[256], size_t *r_size,
     return gpg_error_from_syserror ();
 
   /* variable-length field 1, curve name OID */
-  err = gpg_mpi_write_nohdr (obuf, pkey[0]);
+  err = gpg_mpi_write_opaque_nohdr (obuf, pkey[0]);
   /* fixed-length field 2 */
   iobuf_put (obuf, PUBKEY_ALGO_ECDH);
   /* variable-length field 3, KDF params */
-  err = (err ? err : gpg_mpi_write_nohdr (obuf, pkey[2]));
+  err = (err ? err : gpg_mpi_write_opaque_nohdr (obuf, pkey[2]));
   /* fixed-length field 4 */
   iobuf_write (obuf, "Anonymous Sender    ", 20);
   /* fixed-length field 5, recipient fp (or first 20 octets of fp) */
@@ -192,6 +189,28 @@ derive_kek (size_t kek_size,
             const unsigned char *kdf_params, size_t kdf_params_size)
 {
   gpg_error_t err;
+#if 0 /* GCRYPT_VERSION_NUMBER >= 0x010b00 */
+  /*
+   * Experimental: We will remove this if/endif-conditional
+   * compilation when we update NEED_LIBGCRYPT_VERSION to 1.11.0.
+   */
+  gcry_kdf_hd_t hd;
+  unsigned long param[1];
+
+  param[0] = kek_size;
+  err = gcry_kdf_open (&hd, GCRY_KDF_ONESTEP_KDF, kdf_hash_algo,
+                       param, 1,
+                       secret_x, secret_x_size, NULL, 0, NULL, 0,
+                       kdf_params, kdf_params_size);
+  if (!err)
+    {
+      gcry_kdf_compute (hd, NULL);
+      gcry_kdf_final (hd, kek_size, secret_x);
+      gcry_kdf_close (hd);
+      /* Clean the tail before returning.  */
+      memset (secret_x+kek_size, 0, secret_x_size - kek_size);
+    }
+#else
   gcry_md_hd_t h;
 
   log_assert( gcry_md_get_algo_dlen (kdf_hash_algo) >= 32 );
@@ -211,6 +230,7 @@ derive_kek (size_t kek_size,
   gcry_md_close (h);
   /* Clean the tail before returning.  */
   memset (secret_x+kek_size, 0, secret_x_size - kek_size);
+#endif
   if (DBG_CRYPTO)
     log_printhex (secret_x, kek_size, "ecdh KEK is:");
   return err;
@@ -504,8 +524,7 @@ pk_ecdh_decrypt (gcry_mpi_t *r_result, const byte sk_fp[MAX_FINGERPRINT_LEN],
   size_t nbytes;
   byte *data_buf;
   int data_buf_size;
-  byte *in;
-  const void *p;
+  const unsigned char *p;
   unsigned int nbits;
 
   *r_result = NULL;
@@ -518,7 +537,7 @@ pk_ecdh_decrypt (gcry_mpi_t *r_result, const byte sk_fp[MAX_FINGERPRINT_LEN],
   nbytes = (nbits+7)/8;
 
   data_buf_size = nbytes;
-  if ((data_buf_size & 7) != 1)
+  if ((data_buf_size & 7) != 1 || data_buf_size <= 1 + 8)
     {
       log_error ("can't use a shared secret of %d bytes for ecdh\n",
                  data_buf_size);
@@ -526,7 +545,10 @@ pk_ecdh_decrypt (gcry_mpi_t *r_result, const byte sk_fp[MAX_FINGERPRINT_LEN],
       return gpg_error (GPG_ERR_BAD_DATA);
     }
 
-  data_buf = xtrymalloc_secure( 1 + 2*data_buf_size + 8);
+  /* The first octet is for length.  It's longer than the result
+     because of one additional block of AESWRAP.   */
+  data_buf_size -= 1 + 8;
+  data_buf = xtrymalloc_secure (data_buf_size);
   if (!data_buf)
     {
       err = gpg_error_from_syserror ();
@@ -540,22 +562,18 @@ pk_ecdh_decrypt (gcry_mpi_t *r_result, const byte sk_fp[MAX_FINGERPRINT_LEN],
       gcry_cipher_close (hd);
       return gpg_error (GPG_ERR_BAD_MPI);
     }
-  memcpy (data_buf, p, nbytes);
-  if (data_buf[0] != nbytes-1)
+  if (p[0] != nbytes-1)
     {
       log_error ("ecdh inconsistent size\n");
       xfree (data_buf);
       gcry_cipher_close (hd);
       return gpg_error (GPG_ERR_BAD_MPI);
     }
-  in = data_buf+data_buf_size;
-  data_buf_size = data_buf[0];
 
   if (DBG_CRYPTO)
-    log_printhex (data_buf+1, data_buf_size, "ecdh decrypting :");
+    log_printhex (p+1, nbytes-1, "ecdh decrypting :");
 
-  err = gcry_cipher_decrypt (hd, in, data_buf_size, data_buf+1,
-                             data_buf_size);
+  err = gcry_cipher_decrypt (hd, data_buf, data_buf_size, p+1, nbytes-1);
   gcry_cipher_close (hd);
   if (err)
     {
@@ -565,10 +583,8 @@ pk_ecdh_decrypt (gcry_mpi_t *r_result, const byte sk_fp[MAX_FINGERPRINT_LEN],
       return err;
     }
 
-  data_buf_size -= 8;
-
   if (DBG_CRYPTO)
-    log_printhex (in, data_buf_size, "ecdh decrypted to :");
+    log_printhex (data_buf, data_buf_size, "ecdh decrypted to :");
 
   /* Padding is removed later.  */
   /* if (in[data_buf_size-1] > 8 ) */
@@ -578,7 +594,8 @@ pk_ecdh_decrypt (gcry_mpi_t *r_result, const byte sk_fp[MAX_FINGERPRINT_LEN],
   /*     return gpg_error (GPG_ERR_BAD_KEY); */
   /*   } */
 
-  err = gcry_mpi_scan (r_result, GCRYMPI_FMT_USG, in, data_buf_size, NULL);
+  err = gcry_mpi_scan (r_result, GCRYMPI_FMT_USG, data_buf,
+                       data_buf_size, NULL);
   xfree (data_buf);
   if (err)
     {
